# pio_xounter_sm8_3T.py - 12.09.2021
#
# Counter für bis zu 8 Statemachines bei 3 F_PIO-Takten pro Count
# Die kleinste Zeiteinheit für die Messung ist 3/F_PIO = 25ns bei 120 MHz
#
# Es werden bis zu 64 Pulse-Pause-Zeiten in den Fifo's der Statemachines
# gespeichert.
#
# Der Ruhepegel ist high.
# kann zum Testen an TX einer Serielle Schnittstelle angeschlossen werden.

import time
import uarray as arr
from rp2 import PIO, StateMachine, asm_pio
from machine import Pin

# -  Counter ----------------------------------------
@asm_pio(in_shiftdir=PIO.SHIFT_LEFT, fifo_join=PIO.JOIN_RX )
def pulse_counter():
                 
    in_(null, 32)          #  1 16-Bit DownCounter 0x0000_FFFE vorbereiten
    mov(osr, invert(null)) #  2 0xFFFF_FFFF in OSR
    in_(osr, 15)           #  3 0x0000_EFFF in ISR
    in_(null, 1)           #  4 ox0000:FFFE in ISR
    mov(osr, isr)          #  5 ISR nach OSR kopieren und
    mov(x, osr)            #  6 OSR nach x kopieren
                           
                           #    Damit jede Statemachine ihren Start kennt
                           #    wird das 'y'-Register individuell von aussen beschrieben
                           #    Hier wird auf den Einsatz gewartet ...
    jmp(not_y, "start")    #  7 wenn y = 0, dann direkt weiter zum Start
    label("wait")          #  8 sonst y-Takte warten bis zum Start des Counters
    wait(0, pin, 0)        #  9
    wait(1, pin, 0)        # 10
    jmp(y_dec, "wait")     # 11
       
    label("start")
    
    wait(0, pin, 0)    [5] # 12 Ausgangslevel ist high, jetzt
                           #    warten auf eine fallende Flanke
                           #    1 + [2] Wartetakte dafür, dass der Zäher beeits auf 1 steht
    wrap_target()
                           #    Wait() [2] = 3
                           #    decrementieren, x>0, pin= low, weiter : jmp(x_dec), jmp(pin) jmp(low) = 3 
                           #    decrementieren, x>0, pin=high,   push : jmp(x_dec), jmp(pin), in_(), mow()[2] = 6
                           #    decrementieren, x=0, pin= low, weiter : jmp(x_dec), jmo(pin), jmo(low) = 3  
    label("low")
    jmp(x_dec, "low_1")    # 13 decrementieren unhd immer im nächsten Step weiter
                           
    label("low_1")
    jmp(pin, "low_end")    # 14 der Pin ist high, speichern und weiter nach high
    jmp("low")             # 15 der Pin ist weiter low
    
    label("low_end")
    in_(x, 16)             # 16
    mov(x, osr)        [2] # 17
                           #    decrementieren x>0, pin=high, weiter    jmp(x_dec), jmp(pin), jmp(): 3      
                           #    decrementieren x>0, pin= low, speichern jmp(x_dec), jmp(pin), jmp(end), in_, mov, push : 6
                           #    decrementieren x=0, pin=high, ende      jmp(x_dec), jmp(end) in_, push, jmp()  : 5            
    label("high")
    jmp(x_dec, "high_1")   # 18 Zähler decrementieren, bei 0 beenden
    jmp("train_end")       # 19 speichern, beenden, auf neuen Start warten
    
    label("high_1")        #    hier gehts unabhängig vom Counter Weiter 
       
    jmp(pin, "high_2")     # 20 der Pin ist weiter high
    jmp("high_end")        # 21 der Pin ist low, speichern und weiter nach low
    
    label("high_2")
    jmp("high")            # 21 Wartetakt einlegen
    
    label("high_end")
    in_(x, 16)             # 22 16 Bit speichern, x ist nicht invertiert
    mov(x, osr)            # 23 Counter neu laden und weiter in low
    push()                 # 24
    wrap()
    
    label("train_end")     # 25 der Counter ist bei high-level
    in_(x, 16)             # 26 abgelaufen - wird als Ende interpretiert
    push()                 # 27
    jmp("start")           # 28


class Counter(object):
    """
    Zählt die Dauer von Pulse und Pause in der Einheit 3/counter_freq
    Es werden maximal 64 Werte in den Fifos der 8 Statemachines gespeichert.
    Das X- und Y-Register der Statemachines werden während der Initialisierung
    gesetzt.
    """
    def __init__(self, rx_pin, counter_freq=3000000, statemachines=7):
        self.sm = [""] * statemachines
        self.F_PIO = counter_freq
        self.statemachines = statemachines
        for i in range(self.statemachines):
            self.sm[i] = StateMachine(i, pulse_counter, freq=counter_freq,
                in_base=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
                jmp_pin=Pin(rx_pin, Pin.IN, Pin.PULL_UP) )
            
        self.config_sm()
            
    def config_sm(self):  
        for i in range(self.statemachines):        
            # Prescaler für das Register y setzen, damit das Zählen erst
            # nach dem y'ten. Takt erfolgt.
            # Y "manuell" in die Statemachine ubertragen
            y = 8 * i - 1
            y = 0 if y < 0 else y
            y1 = (y & 0b11110000) // 16
            y2 = (y & 0b00011111)
            
            # Zähler für zu überspringende Takte als 2 Nibble nach y
            self.sm[i].exec("set(y, {})".format(y1))
            self.sm[i].exec("in_(y, 4)")
            self.sm[i].exec("set(y, {})".format(y2))
            self.sm[i].exec("in_(y, 4)")
            self.sm[i].exec("mov(y, isr)")
            self.sm[i].exec("mov(isr, null)")
            
        self.activate(1)
        print("Waiting for Data - Ctrl-C to end")
        print("F_PIO: {} Hz, resolution: {} Nanosec.\n".format(self.F_PIO, 3 * 10**9 / self.F_PIO))
    
    def flush(self):
        # Durch das blockende push() steht u.U. noch ein 9. Datensatz in den
        # Fifos, der muss entsorgt werden 
        for i in range(self.statemachines):
            while self.sm[i].rx_fifo():
                self.sm[i].get()
        
            
    def activate(self, state=1):
        for i in range(self.statemachines):
            self.sm[i].active(1 if state else 0)
    
    def recv(self):
        # liefert zurück, ob die erste Staemachine bereits Daten empfangen hat
        # dann sollten binnen kurzer Zeit auch die anderen gefüllt sein.
        return self.sm[0].rx_fifo()
       
    def print(self, data=None):
        cnt = 0
        idx = 0
        row = 0
        
        for i in range(self.statemachines):        
            x = self.sm[i].rx_fifo()
            cnt += x
            
            if x:
                #print("sm{}: ".format(str(self.sm[i])[12:]), end = " ")
                for _ in range(x):
                    buffer = self.sm[i].get()
                    pulse = 0xFFFF - (buffer // 256**2)
                    pause = 0xFFFF - (buffer & 0xFFFF)
                    
                    # falls die Daten noch benötigt werden,
                    # muss ein Array von 128 Byte übergeben werden
                    if data is not None:
                        data[idx] = pulse
                        idx +=1
                        data[idx] = pause
                        idx += 1
                    #print("{:5}-{:5}".format( pulse, pause), end = " ")
                    if not data:
                        print("#  {:02}  {:6} - {:6}".format(row, pulse, pause))
                    else:
                        print(data[:row * 2])
                    row += 1
            else:
                pass
                
        print("received: {} words\n".format(cnt))
        return cnt * 2


if __name__ == "__main__":
    
    # --- Einstellungen ---------------------------------
    
    # Anzahl der einzusetzenden Statemachines, beginnend mit sm 0
    STATEMACHINES = 8
    F_PIO = 30_000_000    # vorzugsweise mit Teiler 6
    rx_pin = 17           

    # -------------------------------------------------
    
    sm = Counter(rx_pin = rx_pin, counter_freq=F_PIO, statemachines=STATEMACHINES) 
    sm.activate(1)
    
    # Bei Bedarf zur Auswertung die Daten in ein Array kopieren ...
    data = None
    #data = arr.array("B", [0] * 128)
    
    try:
        while True:
            time.sleep(.1)
            # wenn Daten in der 1. Statemachine, dann nur noch kurz warten
            # und anschließend beenden
            if sm.recv():
                time.sleep(.5)
                # print("Words received: ", sm.recv()) 
                # cnt wird nur für die Ausgabe des Arrays "data" benätigt
                cnt = sm.print(data=data)
                sm.config_sm()       
                sm.flush()        # als letztes ausführen !
      
    except KeyboardInterrupt:
        pass
    
    sm.activate(0)
   
    if data:
        print(data[:cnt])
