# pio_uart_rx_4a.py - 12.09.2021
#
# Prüft das Startbit 2x, an der Stelle des Stopbits muss der Pegel high sein.
# Das Ende der Übertragung wird nicht erkannt.
#
# Jeweils nach dem Empfang von 4 Bytes wird ein Interrupt ausgelöst,
# der Interrupt-Handler incrementiert dann einen Zähler für die eingangenen
# Fifo-Words.
#
# Eine große Einschränkung dieser Variante:
# Die Anzahl der übertragenen Byes muss immer durch 4 teilbar sein, sonst
# "stecken" noch Bytes im Fifo. 


from machine import Pin, lightsleep
import uarray as arr
from rp2 import PIO, StateMachine, asm_pio
import time

# ----------------------------------------------------------------------------
#
# Diese Version prüft das Startbit 2x, das Stopbit und die nachfolgende Pause
# Jedes Bit hat eine Länge von 15 Takten.
# Die Datenbits werden jeweils 1x in der Bitmitte abgetastet.
# Takt = Baud * 15   
@rp2.asm_pio(
        #sideset_init=rp2.PIO.OUT_HIGH, 
        in_shiftdir=rp2.PIO.SHIFT_RIGHT,
        fifo_join=rp2.PIO.JOIN_RX ) 

def uart_rx():
    label("restart")  
    mov(isr, null)            #  1 Bei einem Fehler und bei Übertragungsende
                              #    zur Sicherheit das ISR zurücksetzen
    
    wait(0, pin, 0)           #  2 Warten auf eine fallende Flanke
    set(x, 3)                 #  3 Zähler für die 4 Bytes pro Fifo

    label("start")    
    wrap_target()
    set(y, 7)                 #  4 Takt #2 -den Counter fpr die Bits setzen
                              #    Das Signal ist mindestens 4 Takte gesetzt
    jmp(pin, "restart")   [6] #  5 - Fehler, wenn der Takt das Startbit nicht mehr low 
    jmp(pin, "restart")   [5] #  6 - Fehler, wenn der Takt das Startbit nicht mehr low 
    nop()                 [6] #  7 warten bis die Bit-Mitte des 1. Bits erreicht ist

    label("next_bit")
    in_(pins, 1)          [6] #  8 Den Pegel einlesen und nach
    nop()   [6]               #  9
    jmp(y_dec, "next_bit")    # 10 weiteren 15 Takten das nächste Bit einlesen
    
    jmp(pin, "stop_bit_1")    # 11 Das Stopbit muss high, sein 
    jmp("restart")            # 12 wenn nicht gesetzt, dann ist das ein Fehler
    
    label("stop_bit_1")
    jmp(pin, "byte_end")      # 13 Das Stopbit muss high sein, 
    jmp("restart")            # 14 wenn nicht gesetzt, dann ist das ein Fehler                         
    
    label("byte_end")
                   
    jmp(x_dec, "end_1")       # 15 erst nach 4 Bytes pushen

    push()                    # 16 die empfangenen Bits ins Fifo
    set(x, 3)                 # 17 den Zähler für die Bytes neu setzen
    irq(rel(0))               # 18 Interrupt auslösen nach jeweils 4 Byte   
    
    label("end_1")
    set(y, 4)                 # 19 Wartezeit nach dem Stopbit nach 'y'
                              #    wenn innerhalb dieser Zeit der Level nicht
                              #    auf gnd geht, dann ist die Übertragung beendet
    label("end_2")
    jmp(y_dec, "end_3")       # 20 Zähler Wartezeit decrementieren
    jmp("restart")            # 21
    
    label("end_3")            #
    jmp(pin, "end_2")         # 22 hier prüfen, ob eine fallende Flanke auftrat 
    wrap()                    #    wenn ja, dann das nächste Byte einlesen
    
    
class UART_RX(object):
    def __init__(self, statemachine, rx_pin, baud=9600, read_4_bytes=False):
        self.sm = StateMachine(statemachine, uart_rx, 
                in_base=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
                jmp_pin=Pin(rx_pin, Pin.IN, Pin.PULL_UP),
                freq=baud * 15)
        
        self.rx_byte_buffer = arr.array("B", [0] * 4) 
        self.rx_cnt = 0
        self.sm.restart()
        self.sm.irq(self.irq_handler)

    def activate(self, state=1):
        self.sm.active(1 if state else 0)
    
    def irq_handler(self, _):
        # Zählt nur einen Counter für empfangene Words hoch
        # Der Counter muss getestet werden und es müssen
        # die Fifos rechtzeitig abgeholt werden, bevor push() blockt.
        self.rx_cnt += 1
    
    def received(self):
        return self.rx_cnt
            
    def get_data(self):
        # Hier wird genau ein Fifo-Word byteweise zurückgegeben
        temp = self.sm.get()
        self.rx_cnt -= 1
        for i in range(4):
            self.rx_byte_buffer[i] = temp & 0xFF
            temp  >>= 8
        return self.rx_byte_buffer
    
if __name__ == "__main__":
    
    # -- Einstellung ------------------------
    
    BAUD = 9600  
    rx_pin = 17        # input verbinden

# ----------------------------------------------

    rx = UART_RX(statemachine=1, rx_pin=rx_pin, baud=BAUD)
    rx.activate()
    
    # hier hin werden die empfangen Bytes kopiert
    rx_buffer = arr.array("B", [0] * 200)
    rx_idx = 0

    print("Waiting for UART-Data ... Ctrl-C to stop + print.\n")
    
    try:
        while True:
            # lightsleep(10)
            while rx.received():
                buffer = rx.get_data()
        
                for i in range(4):
                    rx_buffer[rx_idx] = buffer[i]
                    rx_idx += 1
               
    except KeyboardInterrupt:
        print(rx_buffer[:rx_idx].decode())
        print("Bytes received: ", rx_idx)
             
    rx.activate(0)
    print("\ndone")
